##############################################################################
# plugin.py - disper plugin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import logging

class Plugin:
    def __init__(self, disper):
        self.disper = disper
        self.log = logging.getLogger('disper.plugin.'+self.__class__.__name__.lower())

# vim:ts=4:sw=4:expandtab:
